/*-------------------------------------------------------------------
Name......... streamsnd.c
Author....... Mike Ballinger
Date......... 07/08/03
Description.. Audio device driver to stream PCM WAV files from, for
..............example, a FLASH card.
..............Supported WAV file formats are (all PCM Uncompressed):
..............     32kHz, 16-bit, Stereo
..............     32kHz, 16-bit, Mono
..............     16kHz, 16-bit, Stereo
..............     16kHz, 16-bit, Mono
..............Up to 4 samples can be playing simultaneously as 4
..............audio channels are available.
-------------------------------------------------------------------*/

/*-------------------------------------------------------------------
DECLARATION FORMATS
Type defs........... UPPER CASE
#defines............ UPPER CASE
Global Functions.... GlobalFunction();
Static Functions.... staticfunction();
Global Variables.... GlobalVariable;
Static Variables.... staticvariable;
Local Variables..... localvariable;
Structure Elements.. Structure.element;
-------------------------------------------------------------------*/

/****	Include Files */

#include "hardware.h"
#include "fpga.h"
#include "fpga.p"
#include "interface.h"

#include "streamsnd.sp"

/****	Externals */
/* 10ms interrupt external */
extern	void (*Int10msLst[])();

/* DMA interrupt external */
extern	void (*IntDMALst[])();

extern	UBYTE	HSFDriverInstalled;

static volatile SNDSTATE	fillstate = SNDSTATE_CHANNEL_LEFT0;

static volatile BYTE	DoneInitStreamAudio=FALSE;
static volatile BYTE	In10ms=FALSE;


/* Output volume register */
static BYTE sfx_vol_register=0;

/* Contains two DMA buffers used for dumping AUDIO data to DMA.*/
static WAVDMAVARS wavDmaVars;

/* We can have up to NUM_CHANNELS WAV files open and playing simultaneously */

static WAVFILEINFO wavFileChannel[NUM_CHANNELS];


/* Current output DEVICE. This is initialised when the function InitStreamedAudio() is called */
const DEVICE * currentDev;


/* average of two signed words */
#define SIGNEDWORDAVERAGE(x,y)  (WORD)((LONG)(((LONG)(x)+(LONG)(y))/2))



/*-------------------------------------------------------------------
Look-up table for U-law decompress

Generated using the following decompression algorithm:
Reference: sox.exe source code (sound processing utility)

#define QUANT_MASK	(0x0f)
#define BIAS		(0x84)
#define SEG_MASK	(0x70)
#define SEG_SHIFT	(4)
#define SIGN_BIT	(0x80)

short ulaw2linear(unsigned char u_val)
{
	int t;

	u_val = ~u_val;
	t = ((u_val & QUANT_MASK) <<3) + BIAS;
	t <<= ((unsigned)u_val & SEG_MASK) >> SEG_SHIFT;
	return( (u_val&SIGN_BIT) ? (BIAS-t) : (t-BIAS));
}

-------------------------------------------------------------------*/
static const WORD ulaw_lookup[ 256 ] =
{
    0x8284, 0x8684, 0x8a84, 0x8e84, 0x9284, 0x9684, 0x9a84, 0x9e84,
    0xa284, 0xa684, 0xaa84, 0xae84, 0xb284, 0xb684, 0xba84, 0xbe84,
    0xc184, 0xc384, 0xc584, 0xc784, 0xc984, 0xcb84, 0xcd84, 0xcf84,
    0xd184, 0xd384, 0xd584, 0xd784, 0xd984, 0xdb84, 0xdd84, 0xdf84,
    0xe104, 0xe204, 0xe304, 0xe404, 0xe504, 0xe604, 0xe704, 0xe804,
    0xe904, 0xea04, 0xeb04, 0xec04, 0xed04, 0xee04, 0xef04, 0xf004,
    0xf0c4, 0xf144, 0xf1c4, 0xf244, 0xf2c4, 0xf344, 0xf3c4, 0xf444,
    0xf4c4, 0xf544, 0xf5c4, 0xf644, 0xf6c4, 0xf744, 0xf7c4, 0xf844,
    0xf8a4, 0xf8e4, 0xf924, 0xf964, 0xf9a4, 0xf9e4, 0xfa24, 0xfa64,
    0xfaa4, 0xfae4, 0xfb24, 0xfb64, 0xfba4, 0xfbe4, 0xfc24, 0xfc64,
    0xfc94, 0xfcb4, 0xfcd4, 0xfcf4, 0xfd14, 0xfd34, 0xfd54, 0xfd74,
    0xfd94, 0xfdb4, 0xfdd4, 0xfdf4, 0xfe14, 0xfe34, 0xfe54, 0xfe74,
    0xfe8c, 0xfe9c, 0xfeac, 0xfebc, 0xfecc, 0xfedc, 0xfeec, 0xfefc,
    0xff0c, 0xff1c, 0xff2c, 0xff3c, 0xff4c, 0xff5c, 0xff6c, 0xff7c,
    0xff88, 0xff90, 0xff98, 0xffa0, 0xffa8, 0xffb0, 0xffb8, 0xffc0,
    0xffc8, 0xffd0, 0xffd8, 0xffe0, 0xffe8, 0xfff0, 0xfff8, 0x0000,
    0x7d7c, 0x797c, 0x757c, 0x717c, 0x6d7c, 0x697c, 0x657c, 0x617c,
    0x5d7c, 0x597c, 0x557c, 0x517c, 0x4d7c, 0x497c, 0x457c, 0x417c,
    0x3e7c, 0x3c7c, 0x3a7c, 0x387c, 0x367c, 0x347c, 0x327c, 0x307c,
    0x2e7c, 0x2c7c, 0x2a7c, 0x287c, 0x267c, 0x247c, 0x227c, 0x207c,
    0x1efc, 0x1dfc, 0x1cfc, 0x1bfc, 0x1afc, 0x19fc, 0x18fc, 0x17fc,
    0x16fc, 0x15fc, 0x14fc, 0x13fc, 0x12fc, 0x11fc, 0x10fc, 0x0ffc,
    0x0f3c, 0x0ebc, 0x0e3c, 0x0dbc, 0x0d3c, 0x0cbc, 0x0c3c, 0x0bbc,
    0x0b3c, 0x0abc, 0x0a3c, 0x09bc, 0x093c, 0x08bc, 0x083c, 0x07bc,
    0x075c, 0x071c, 0x06dc, 0x069c, 0x065c, 0x061c, 0x05dc, 0x059c,
    0x055c, 0x051c, 0x04dc, 0x049c, 0x045c, 0x041c, 0x03dc, 0x039c,
    0x036c, 0x034c, 0x032c, 0x030c, 0x02ec, 0x02cc, 0x02ac, 0x028c,
    0x026c, 0x024c, 0x022c, 0x020c, 0x01ec, 0x01cc, 0x01ac, 0x018c,
    0x0174, 0x0164, 0x0154, 0x0144, 0x0134, 0x0124, 0x0114, 0x0104,
    0x00f4, 0x00e4, 0x00d4, 0x00c4, 0x00b4, 0x00a4, 0x0094, 0x0084,
    0x0078, 0x0070, 0x0068, 0x0060, 0x0058, 0x0050, 0x0048, 0x0040,
    0x0038, 0x0030, 0x0028, 0x0020, 0x0018, 0x0010, 0x0008, 0x0000
};

/*-------------------------------------------------------------------
Look-up table for A-law decompress

Generated using the following decompression algorithm:
Reference: sox.exe source code (sound processing utility)

#define QUANT_MASK	(0x0f)
#define BIAS		(0x84)
#define SEG_MASK	(0x70)
#define SEG_SHIFT	(4)
#define SIGN_BIT	(0x80)

short alaw2linear(unsigned char a_val)
{
	int		t;
	int		seg;

	a_val ^= 0x55;
	t = (a_val & QUANT_MASK) << 4;
	seg = ((unsigned)a_val & SEG_MASK) >> SEG_SHIFT;
	switch (seg) {
		case 0:
			t += 8;
			break;
		case 1:
			t += 0x108;
			break;
		default:
			t += 0x108;
			t <<= seg - 1;
	}
	return ((a_val & SIGN_BIT) ? t : -t);
}


-------------------------------------------------------------------*/
static const WORD alaw_lookup[ 256 ] =
{
    0xea80, 0xeb80, 0xe880, 0xe980, 0xee80, 0xef80, 0xec80, 0xed80,
    0xe280, 0xe380, 0xe080, 0xe180, 0xe680, 0xe780, 0xe480, 0xe580,
    0xf540, 0xf5c0, 0xf440, 0xf4c0, 0xf740, 0xf7c0, 0xf640, 0xf6c0,
    0xf140, 0xf1c0, 0xf040, 0xf0c0, 0xf340, 0xf3c0, 0xf240, 0xf2c0,
    0xaa00, 0xae00, 0xa200, 0xa600, 0xba00, 0xbe00, 0xb200, 0xb600,
    0x8a00, 0x8e00, 0x8200, 0x8600, 0x9a00, 0x9e00, 0x9200, 0x9600,
    0xd500, 0xd700, 0xd100, 0xd300, 0xdd00, 0xdf00, 0xd900, 0xdb00,
    0xc500, 0xc700, 0xc100, 0xc300, 0xcd00, 0xcf00, 0xc900, 0xcb00,
    0xfea8, 0xfeb8, 0xfe88, 0xfe98, 0xfee8, 0xfef8, 0xfec8, 0xfed8,
    0xfe28, 0xfe38, 0xfe08, 0xfe18, 0xfe68, 0xfe78, 0xfe48, 0xfe58,
    0xffa8, 0xffb8, 0xff88, 0xff98, 0xffe8, 0xfff8, 0xffc8, 0xffd8,
    0xff28, 0xff38, 0xff08, 0xff18, 0xff68, 0xff78, 0xff48, 0xff58,
    0xfaa0, 0xfae0, 0xfa20, 0xfa60, 0xfba0, 0xfbe0, 0xfb20, 0xfb60,
    0xf8a0, 0xf8e0, 0xf820, 0xf860, 0xf9a0, 0xf9e0, 0xf920, 0xf960,
    0xfd50, 0xfd70, 0xfd10, 0xfd30, 0xfdd0, 0xfdf0, 0xfd90, 0xfdb0,
    0xfc50, 0xfc70, 0xfc10, 0xfc30, 0xfcd0, 0xfcf0, 0xfc90, 0xfcb0,
    0x1580, 0x1480, 0x1780, 0x1680, 0x1180, 0x1080, 0x1380, 0x1280,
    0x1d80, 0x1c80, 0x1f80, 0x1e80, 0x1980, 0x1880, 0x1b80, 0x1a80,
    0x0ac0, 0x0a40, 0x0bc0, 0x0b40, 0x08c0, 0x0840, 0x09c0, 0x0940,
    0x0ec0, 0x0e40, 0x0fc0, 0x0f40, 0x0cc0, 0x0c40, 0x0dc0, 0x0d40,
    0x5600, 0x5200, 0x5e00, 0x5a00, 0x4600, 0x4200, 0x4e00, 0x4a00,
    0x7600, 0x7200, 0x7e00, 0x7a00, 0x6600, 0x6200, 0x6e00, 0x6a00,
    0x2b00, 0x2900, 0x2f00, 0x2d00, 0x2300, 0x2100, 0x2700, 0x2500,
    0x3b00, 0x3900, 0x3f00, 0x3d00, 0x3300, 0x3100, 0x3700, 0x3500,
    0x0158, 0x0148, 0x0178, 0x0168, 0x0118, 0x0108, 0x0138, 0x0128,
    0x01d8, 0x01c8, 0x01f8, 0x01e8, 0x0198, 0x0188, 0x01b8, 0x01a8,
    0x0058, 0x0048, 0x0078, 0x0068, 0x0018, 0x0008, 0x0038, 0x0028,
    0x00d8, 0x00c8, 0x00f8, 0x00e8, 0x0098, 0x0088, 0x00b8, 0x00a8,
    0x0560, 0x0520, 0x05e0, 0x05a0, 0x0460, 0x0420, 0x04e0, 0x04a0,
    0x0760, 0x0720, 0x07e0, 0x07a0, 0x0660, 0x0620, 0x06e0, 0x06a0,
    0x02b0, 0x0290, 0x02f0, 0x02d0, 0x0230, 0x0210, 0x0270, 0x0250,
    0x03b0, 0x0390, 0x03f0, 0x03d0, 0x0330, 0x0310, 0x0370, 0x0350
};



/**********************************************************************************************/
/*                                                                                            */
/* API Functions available to end user                                                        */
/*                                                                                            */
/**********************************************************************************************/

/*-------------------------------------------------------------------
Name......... InitStreamedAudio()
Author....... Mark Perkins
Date......... 07/03/05
Description.. Initialises the streamed audio driver. This must be
............. called before any other streamed audio functions
Parameters... dev: output device
Return....... 0 On Success
............. Error code on error
History...... 05/11/03 - Initial version  - MJB
............. 03/05/05 - Rewrite MP -- Clear wav info structures.
-------------------------------------------------------------------*/
BYTE  InitStreamedAudio( const DEVICE* dev )
{
	LONG i;

	/* don't setup this driver if the HSF driver is installed */
	if(HSFDriverInstalled)
	{
		return(INITERR);
	}

	/* Dont let it be called more than once */
	if (DoneInitStreamAudio)
	{
		return(INITERR);
	}

	/* To begin with, no DMA transfer is in progress */
	wavDmaVars.activeDMAbuffer = FALSE;

	/* Reset DMA variables */
	wavDmaVars.dmaRequired[0] = TRUE;
	wavDmaVars.dmaRequired[1] = FALSE;

    wavDmaVars.bufferready[0] = FALSE;
    wavDmaVars.bufferready[1] = FALSE;

	wavDmaVars.bufferUsed[0] = 0;
	wavDmaVars.bufferUsed[1] = 0;

   	/* Set next buffer to fill */
	wavDmaVars.nextBufferToFill = 0;

	/* Must make a global copy of the current output device
	   for use in other functions */
	currentDev = dev;

	/* No WAV files are currently active so reset all variables*/
	for ( i = 0; i < NUM_CHANNELS; i ++ )
    {
		ClearAudioChannel(i);
    }

	/* install sound into 10ms interupt list */
	InstallIntFnc( SFX10msInterruptRoutine, Int10msLst );

	/* install DMA isr into DMA interrupt list */
	InstallIntFnc( DMA2InterruptRoutine, IntDMALst );

	/* Enable SFX Audio Bit In FPGA */
	*Xcra |= Sfxenable;

	/* Set volume of both channels to zero */
	SetAudioVolume( AUDIO_STEREO0, 0 );

	/* Set called Flag */
	DoneInitStreamAudio = TRUE;

	return(0);
}

/*-------------------------------------------------------------------
Name......... SearchForWavFile()
Author....... Mark Perkins
Date......... 03/05/05
Description.. Searches for a wav file from a filename and populates
............. the file descriptor. This uses less ram than the
............. InitWavFile function which uses a wavinfo structure.
Parameters... filename: WAV filename, e.g. "music.wav"
............. File descriptor.
Return....... 0 On Success
............. Error code on error
History...... 03/05/05 - MP -- Initial Version.

-------------------------------------------------------------------*/
BYTE SearchForWavFile( BYTE *filename, FILE_DESCRIPTOR *file )
{
	/* Open File */
	if ( FileOpen( currentDev, filename, file, READ_ONLY ) != 0 ) return(0);
	else return(INITERR_FILENOTFOUND);
}

/*-------------------------------------------------------------------
Name......... PlaybackCachedWavFile()
Author....... Mark Perkins
Date......... 03/05/05
Description.. Plays WAV file on required channel. NOTE a call to
............. SearchForWavFile() must be made first to do initial
............. processing on the WAV file.
Parameters... channel: Output audio channel, e.g. AUDIO_RIGHT1
............. File descriptor:
Return....... 0 On Success
............. Error code on error
History...... 03/05/05 - MP -- Initial Version.

-------------------------------------------------------------------*/
BYTE PlaybackWavFile( BYTE channel, FILE_DESCRIPTOR *file)
{
	UBYTE wavChannel;
	BYTE audioChannel;

	/* Obtain audio channel (without AUDIO_LOOPED bit set) */
	audioChannel = channel & ~AUDIO_LOOPED;

	/* First ensure required sound channel is not busy. If it is, return */
	if ( SoundChannelBusy( audioChannel ) )
	{
		return(PLAYERR_CHANNELBUSY);
	}

	switch ( audioChannel )
	{
		case AUDIO_LEFT0:
			wavChannel = CHANNEL_LEFT0;
			break;

		case AUDIO_RIGHT0:
			wavChannel = CHANNEL_RIGHT0;
			break;

		case AUDIO_STEREO0:
			wavChannel = CHANNEL_LEFT0;
			break;

		case AUDIO_LEFT1:
			wavChannel = CHANNEL_LEFT1;
			break;

		case AUDIO_RIGHT1:
			wavChannel = CHANNEL_RIGHT1;
			break;

		case AUDIO_STEREO1:
			wavChannel = CHANNEL_LEFT1;
			break;

		default:
			return(PLAYERR_INVALIDCHANNEL);
			break;
	}

	wavFileChannel[wavChannel].channel = audioChannel;

	if ( channel & AUDIO_LOOPED )
	{
		wavFileChannel[wavChannel].looped = TRUE;
	}
	else
	{
		wavFileChannel[wavChannel].looped = FALSE;
	}

	wavFileChannel[wavChannel].file = *file;

	wavFileChannel[wavChannel].status = CHANNEL_START;

	return(0);
}

/*-------------------------------------------------------------------
Name......... InitWavFile()
Author....... Mike Ballinger
Date......... 05/11/03
Description.. Initialises a WAV file before playback using function
............. PlaybackCachedWavFile().
Parameters... filename: WAV filename, e.g. "music.wav"
............. wavFileInfo: Structure where WAV info will be stored.
Return....... 0 On Success
............. Error code on error
History...... 05/11/03 - Initial version  - MJB
............. 07/03/05 - Rewrite MP -- Dont read wav header here as
............. it will cause a sector read.
-------------------------------------------------------------------*/
BYTE InitWavFile( BYTE *filename, WAVFILEINFO *wavFileInfo )
{
	/* Open File */
	if ( FileOpen( currentDev, filename, &wavFileInfo->file, READ_ONLY ) != 0 )
	{
		return(0);
	}
	else
	{
		return(INITERR_FILENOTFOUND);
	}
}

/*-------------------------------------------------------------------
Name......... PlaybackCachedWavFile()
Author....... Mark Perkins
Date......... 07/03/05
Description.. Plays WAV file on required channel. NOTE a call to
............. InitWavFile() must be made first to do initial
............. processing on the WAV file.
Parameters... channel: Output audio channel, e.g. AUDIO_RIGHT1
............. wavFileInfo: Sturcture containing WAV file info
Return....... 0 On Success
............. Error code on error
History...... 05/11/03 - Initial version  - MJB
............. 07/03/05 - Rewrite MP -- Changed to playing using status.
-------------------------------------------------------------------*/
BYTE PlaybackCachedWavFile( UBYTE channel, WAVFILEINFO *wavFileInfo )
{
	UBYTE wavChannel;
	BYTE audioChannel;
	WAVFILEINFO tmp;

	/* Obtain audio channel (without AUDIO_LOOPED bit set) */
	audioChannel = channel & ~AUDIO_LOOPED;

	/* First ensure required sound channel is not busy. If it is, return */
	if ( SoundChannelBusy( audioChannel ) )
	{
		return(PLAYERR_CHANNELBUSY);
	}


	switch ( audioChannel )
	{
		case AUDIO_LEFT0:
			wavChannel = CHANNEL_LEFT0;
			break;

		case AUDIO_RIGHT0:
			wavChannel = CHANNEL_RIGHT0;
			break;

		case AUDIO_STEREO0:
			wavChannel = CHANNEL_LEFT0;
			break;

		case AUDIO_LEFT1:
			wavChannel = CHANNEL_LEFT1;
			break;

		case AUDIO_RIGHT1:
			wavChannel = CHANNEL_RIGHT1;
			break;

		case AUDIO_STEREO1:
			wavChannel = CHANNEL_LEFT1;
			break;

		default:
			return(PLAYERR_INVALIDCHANNEL);
			break;

	}

	wavFileChannel[wavChannel].channel = audioChannel;

	if ( channel & AUDIO_LOOPED )
	{
		wavFileChannel[wavChannel].looped = TRUE;
	}
	else
	{
		wavFileChannel[wavChannel].looped = FALSE;
	}

	tmp = *wavFileInfo;

	wavFileChannel[wavChannel].file = tmp.file;

	wavFileChannel[wavChannel].status = CHANNEL_START;

	return(0);
}

/*-------------------------------------------------------------------
Name......... GetChannelStatus()
Author....... Mark Perkins
Date......... 07/03/05
Description.. Returns the status of the audio channel audio channel.
Parameters... audioChannel: channel, e.g. AUDIO_STEREO0
Return....... 0 On Success
............. Negative error code on failure.
History...... 07/03/05 - Initial version - MP
-------------------------------------------------------------------*/
CHANNELSTATUS GetChannelStatus(UBYTE audioChannel)
{
    switch ( audioChannel )
    {
        case AUDIO_LEFT0:
			return(wavFileChannel[CHANNEL_LEFT0].status);
			break;

        case AUDIO_RIGHT0:
			return(wavFileChannel[CHANNEL_RIGHT0].status);
			break;

        case AUDIO_STEREO0:
			return(wavFileChannel[CHANNEL_LEFT0].status);
			break;

        case AUDIO_LEFT1:
			return(wavFileChannel[CHANNEL_LEFT1].status);
			break;

        case AUDIO_RIGHT1:
			return(wavFileChannel[CHANNEL_RIGHT1].status);
			break;

        case AUDIO_STEREO1:
			return(wavFileChannel[CHANNEL_LEFT1].status);
			break;

		default:
			return(ERR_INVALIDCHANNEL);
			break;
    }
}

/*-------------------------------------------------------------------
Name......... QuietChannel()
Author....... Mark Perkins
Date......... 07/03/05
Description.. Silences any audio currently playing on the requested
............. channel
Parameters... audioChannel: channel, e.g. AUDIO_STEREO0
Return....... 0 On Success
............. Negative error code on failure.
History...... 18/11/03 - Initial version - MJB
............. 07/03/05 - Rewrite MP -- Call stop channel.
-------------------------------------------------------------------*/
BYTE QuietChannel( UBYTE audioChannel )
{
	BYTE ret;

    ret = StopAudioChannel( audioChannel );

    return(ret);
}

/*-------------------------------------------------------------------
Name......... StopAudioChannel()
Author....... Mark Perkins
Date......... 07/03/05
Description.. Silences any audio currently playing on the requested
............. channel.
.............
Parameters... audioChannel: channel, e.g. AUDIO_STEREO0
Return....... 0 On Success
............. Negative error code on failure.
History...... 19/08/2004 - Initial version - AA
............. 07/03/05 - Rewrite MP -- Changed to stop channel by
............. using the status.
-------------------------------------------------------------------*/
BYTE StopAudioChannel( UBYTE audioChannel )
{
   	switch ( audioChannel )
   	{
   	    case AUDIO_LEFT0:
			ClearAudioChannel(CHANNEL_LEFT0);
			return(0);
   	        break;

   	    case AUDIO_RIGHT0:
			ClearAudioChannel(CHANNEL_RIGHT0);
			return(0);
   	        break;

        case AUDIO_STEREO0:
			ClearAudioChannel(CHANNEL_LEFT0);
			ClearAudioChannel(CHANNEL_RIGHT0);
			return(0);
            break;

        case AUDIO_LEFT1:
			ClearAudioChannel(CHANNEL_LEFT1);
			return(0);
            break;

       	case AUDIO_RIGHT1:
			ClearAudioChannel(CHANNEL_RIGHT1);
			return(0);
            break;

        case AUDIO_STEREO1:
			ClearAudioChannel(CHANNEL_LEFT1);
			ClearAudioChannel(CHANNEL_RIGHT1);
			return(0);
            break;

		default:
			return(ERR_INVALIDCHANNEL);
			break;
	}
}

/*-------------------------------------------------------------------
Name......... SetAudioVolume()
Author....... Mike Ballinger
Date......... 25/11/03
Description.. Sets volume on the requested channel(s)
Parameters... audioChannel: channel - AUDIO_LEFT0, AUDIO_RIGHT0,
.............                         AUDIO_STEREO0
............. volume: 0 (min) - 255 (max)
Return....... None
History...... 25/11/03 - Initial version - MJB
-------------------------------------------------------------------*/
void SetAudioVolume( UBYTE audioChannel, UBYTE volume )
{
	BYTE reqVolume, volLeft, volRight;

	/* Change 256 Levels to 16 */
	reqVolume = ( volume >> 4 ) & 0x0f;

	/* Store Current Volumes */
	volLeft = sfx_vol_register & 0xf0;
	volRight = sfx_vol_register & 0x0f;

	switch ( audioChannel )
	{
		case AUDIO_LEFT0:
			volLeft = ( (reqVolume & 0x0f) * 16 );
			break;

		case AUDIO_RIGHT0:
			volRight = reqVolume & 0x0f;
			break;

		case AUDIO_STEREO0:
			volLeft = ( (reqVolume & 0x0f) * 16 );
			volRight = reqVolume & 0x0f;
			break;

		default:
			break;
	}

	/* Update Volume Register with new value */
	sfx_vol_register = ( volLeft | volRight );
	*Xsfxvol = sfx_vol_register;
}

/*-------------------------------------------------------------------
Name......... SoundChannelBusy()
Author....... Mark Perkins
Date......... 07/03/05
Description.. Checks to see if the specified sound channel is busy
Parameters... audioChannel: Audio Channel to check, e.g. AUDIO_LEFT0
Return....... 0 if channel available, 1 if channel is busy
............. Negative error code on failure.
History...... 05/11/03 - Initial version  - MJB
............. 07/03/05 - Rewrite MP -- Change to test status.
-------------------------------------------------------------------*/
BYTE SoundChannelBusy( UBYTE audioChannel )
{
	switch ( audioChannel )
	{
		case AUDIO_LEFT0:
			/* Return if left channel is already active */
			if ( wavFileChannel[CHANNEL_LEFT0].status !=
				CHANNEL_IDLE && wavFileChannel[CHANNEL_LEFT0].status !=
				CHANNEL_WAVERROR && wavFileChannel[CHANNEL_LEFT0].status
				!= CHANNEL_INVALIDCHANNEL)
			{
				return( 1 );
			}
			break;

		case AUDIO_RIGHT0:
			/* Return if right channel is already active */
			if ( wavFileChannel[CHANNEL_RIGHT0].status !=
				CHANNEL_IDLE && wavFileChannel[CHANNEL_RIGHT0].status !=
				CHANNEL_WAVERROR && wavFileChannel[CHANNEL_RIGHT0].status
				!= CHANNEL_INVALIDCHANNEL)
			{
				return( 1 );
			}

			/* Return if currently playing a stereo sample */
			if ( wavFileChannel[CHANNEL_LEFT0].status !=
				CHANNEL_IDLE && wavFileChannel[CHANNEL_LEFT0].status !=
				CHANNEL_WAVERROR && wavFileChannel[CHANNEL_LEFT0].status
				!= CHANNEL_INVALIDCHANNEL && wavFileChannel[CHANNEL_LEFT0].channel
				== AUDIO_STEREO0)
			{
				return( 1 );
			}

			break;

		case AUDIO_STEREO0:
			/* Return if either left or right channel is currently active */
			if ( wavFileChannel[CHANNEL_LEFT0].status !=
				CHANNEL_IDLE && wavFileChannel[CHANNEL_LEFT0].status !=
				CHANNEL_WAVERROR && wavFileChannel[CHANNEL_LEFT0].status !=
				CHANNEL_INVALIDCHANNEL)
			{
				return( 1 );
			}
			break;

		case AUDIO_LEFT1:
			/* Return if left channel is already active */
			if ( wavFileChannel[CHANNEL_LEFT1].status !=
				CHANNEL_IDLE && wavFileChannel[CHANNEL_LEFT1].status !=
				CHANNEL_WAVERROR && wavFileChannel[CHANNEL_LEFT1].status !=
				CHANNEL_INVALIDCHANNEL)
			{
				return( 1 );
			}
			break;

		case AUDIO_RIGHT1:
			/* Return if right channel is already active */
			if ( wavFileChannel[CHANNEL_RIGHT1].status !=
				CHANNEL_IDLE && wavFileChannel[CHANNEL_RIGHT1].status !=
				CHANNEL_WAVERROR && wavFileChannel[CHANNEL_RIGHT1].status !=
				CHANNEL_INVALIDCHANNEL )
			{
				return( 1 );
			}

			if ( wavFileChannel[CHANNEL_LEFT1].status !=
				CHANNEL_IDLE && wavFileChannel[CHANNEL_LEFT1].status !=
				CHANNEL_WAVERROR && wavFileChannel[CHANNEL_LEFT1].status !=
				CHANNEL_INVALIDCHANNEL && wavFileChannel[CHANNEL_LEFT1].channel
				== AUDIO_STEREO1)
			{
				return( 1 );
			}
			break;

		case AUDIO_STEREO1:
			/* Return if either left or right channel is currently active */
			if ( wavFileChannel[CHANNEL_LEFT1].status !=
				CHANNEL_IDLE && wavFileChannel[CHANNEL_LEFT1].status !=
				CHANNEL_WAVERROR && wavFileChannel[CHANNEL_LEFT1].status !=
				CHANNEL_INVALIDCHANNEL)
			{
				return( 1 );
			}
			break;

		default:
			return(ERR_INVALIDCHANNEL);
			break;
	}

	return(0);
}

/**********************************************************************************************/
/*                                                                                            */
/* Internal module functions - not available to end user                                      */
/*                                                                                            */
/**********************************************************************************************/

/*-------------------------------------------------------------------
Name......... VerifyFileData()
Author....... Mike Ballinger
Date......... 17/11/03
Description.. Verifies that the buffer contains the specified data
Parameters... buffer: the buffer to look in
............. offset: current offset of buffer
............. data: the data to compare against
............. dataLength: number of bytes in data
Return....... 0 if the buffer contains the data, 1 otherwise
..............NOTE: the value in offset is incremented by the
............. specified number of bytes.
History...... 17/11/03 - Initial version  - MJB
-------------------------------------------------------------------*/
static BYTE VerifyFileData( UBYTE *buffer, ULONG *offset, UBYTE *data, UBYTE dataLength )
{
	ULONG count;
	UBYTE result = 0;

	for ( count = 0; count < dataLength; count ++ )
	{
		if ( data[count] != buffer[(*offset)++] )
		{
			result = TRUE;
		}
	}

	return( result );
}

/*-------------------------------------------------------------------
Name......... GetFileNumber()
Author....... Mike Ballinger
Date......... 17/11/03
Description.. Returns a number (the specified number of bytes long)
............. from a buffer at the specified offset.
Parameters... buffer: buffer containing data
............. offset: current buffer offset
............. numBytes: number of bytes in number (1 to 4)
Return....... The number.
..............NOTE: the variable offset is incremented by the
............. specified number of bytes.
History...... 17/11/03 - Initial version - MJB
-------------------------------------------------------------------*/
static ULONG GetFileNumber( UBYTE *buffer, ULONG *offset, ULONG numBytes )
{
	ULONG count;
	UBYTE bitShift = 0;
	ULONG result = 0;

	for ( count = 0; count < numBytes; count ++ )
	{
		result += (LONG)buffer[(*offset)++] << bitShift;
		bitShift += 8;
	}

	return( result );
}

/*-------------------------------------------------------------------
Name......... ZeroDMABuffer()
Author....... Mike Ballinger
Date......... 17/11/03
Description.. Zeros the specified DMA buffer.
Parameters... dmaChannel: DMA buffer (either 0 or 1)
Return....... None
History...... 11/12/03 - Optimised for speed  - MJB
-------------------------------------------------------------------*/
static void ZeroDMABuffer( UBYTE dmaChannel )
{
	register ULONG offset;
	WORD *localDMABuffer;


	if (dmaChannel == 0 || dmaChannel == 1)
	{
		localDMABuffer = &wavDmaVars.buffer[dmaChannel][0];

		offset = WAV_DMA_BUFFER_SIZE;
		while ( offset )
		{
			*localDMABuffer++ = 0;
			offset --;
		}
	}
}

/*-------------------------------------------------------------------
Name......... FillDMASample()
Author....... Mike Ballinger
Date......... 17/11/03
Description.. Fills the specified DMA buffer with data from the
............. specified mono WAV file.
Parameters... dmaChannel: DMA buffer to fill - either 0 or 1
............. wavFileInfo: WAV file information stored within
............. structure of type WAVFILEINFO
............. buffer: Buffer containing audio data from WAV file
............. bufferUsed: Number of samples in buffer
Return....... None
History...... 11/12/03 - Optimised for speed - MJB
............. 26/07/04 - Added (AA):
............. byte swapping (endian conversion) removed to calling function
............. 13/04/05 - MP - Changed to one fill function.
-------------------------------------------------------------------*/
static void FillDMASample( UBYTE dmaChannel, UBYTE channel, WORD *buffer,
						   ULONG bufferUsed )
{
	WORD currentSample, mixSample;
    WORD leftSample, rightSample;

	WORD *localDMABuffer, *localDMABufferStart, *localBuffer;
	ULONG localDMABufferUsed;
	register ULONG i;
	register ULONG offset;
	register ULONG dmaOutputReps;
    UBYTE numofchannels;


	/* default dma output reps to 1 and change to 2 in 32K sample */
	dmaOutputReps = 1;
	if ( wavFileChannel[channel].sampleRate == 16000 )
	{
		dmaOutputReps = 2;
	}

	/* Make a local copy of DMA buffer. This should speed things up. */
	localDMABuffer = &wavDmaVars.buffer[dmaChannel][0];
	localDMABufferStart = localDMABuffer;
	localBuffer = buffer;

    numofchannels = wavFileChannel[channel].numChannels;


    /* Update */
	switch( wavFileChannel[channel].channel )
	{
		case AUDIO_LEFT0:
			offset = bufferUsed;
			while( offset )
			{
				currentSample = *localBuffer++;

				i = dmaOutputReps;
				while( i )
				{
					/* Right audio channel is first on Pluto, whereas left is
                       first in a WAV file. */
					localDMABuffer++; /* Right DMA channel */
					*localDMABuffer++ = currentSample; /* Left DMA channel */
					i --;
				}

				offset --;
			}
			break;

		case AUDIO_RIGHT0:
			offset = bufferUsed;
			while( offset )
			{
				currentSample = *localBuffer++;

				i = dmaOutputReps;
				while( i )
				{
					/* Right audio channel is first on Pluto, whereas left is
                       first in a WAV file. */
					*localDMABuffer++ = currentSample; /* Right DMA channel */
					localDMABuffer++; /* Left DMA channel */
					i --;
				}

				offset --;
			}
			break;

		case AUDIO_STEREO0:
            if (numofchannels == 2)
            {
                i = wavDmaVars.bufferUsed[dmaChannel];
                while(i)
                {
                    *localDMABuffer++;
                    i--;
                }

                offset = bufferUsed >> 1;
                while( offset )
                {
                    /* Must reverse endian on all samples */
                    leftSample = *localBuffer++;
                    rightSample = *localBuffer++;
                    i = dmaOutputReps;
                    while( i )
                    {
                        /* Right audio channel is first on Pluto, whereas left is
                           first in a WAV file.  Must also reverse endian on all
                           samples */
                        *localDMABuffer++ = rightSample; /* Right DMA channel */
                        *localDMABuffer++ = leftSample; /* Left DMA channel */
                        i --;
                    }
                    offset --;
                }
            }
            else
            {
                offset = bufferUsed;
                while( offset )
                {
				    currentSample = *localBuffer++;

				    i = dmaOutputReps;
				    while( i )
				    {
    					/* Right audio channel is first on Pluto, whereas left is
                           first in a WAV file. */
		      			*localDMABuffer++ = currentSample; /* Right DMA channel */
			     		*localDMABuffer++ = currentSample; /* Left DMA channel */
				    	i --;
				    }
                    offset --;
                }
            }
			break;

		/* The following audio channels are mixed */
		case AUDIO_LEFT1:
			offset = bufferUsed;
			while( offset )
			{
				currentSample = *localBuffer++;

				i = dmaOutputReps;
				while( i )
				{
					/* Right audio channel is first on Pluto, whereas left is
                       first in a WAV file. */
					localDMABuffer++; /* Right DMA channel */
					mixSample = SIGNEDWORDAVERAGE(*localDMABuffer, currentSample );
					*localDMABuffer++ = mixSample; /* Left DMA channel */
					i --;
				}

				offset --;
			}
			break;

		case AUDIO_RIGHT1:
			offset = bufferUsed;
			while( offset )
			{
				currentSample = *localBuffer++;

				i = dmaOutputReps;
				while( i )
				{
					/* Right audio channel is first on Pluto, whereas left is
                       first in a WAV file. */
					mixSample = SIGNEDWORDAVERAGE(*localDMABuffer, currentSample );
					*localDMABuffer++ = mixSample; /* Right DMA channel */
					localDMABuffer++; /* Left DMA channel */
					i --;
				}

				offset --;
			}
			break;

		case AUDIO_STEREO1:
            if (numofchannels == 2)
            {
                i = wavDmaVars.bufferUsed[dmaChannel];
                while(i)
                {
                    *localDMABuffer++;
                    i--;
                }

                offset = bufferUsed >> 1;
                while( offset )
                {
                    /* Must reverse endian on all samples */
                    leftSample = *localBuffer++;
                    rightSample = *localBuffer++;

                    i = dmaOutputReps;
                    while( i )
                    {
                        /* Right audio channel is first on Pluto, whereas left is
                        first in a WAV file.  Must also reverse endian on all
                        samples */
                        mixSample = SIGNEDWORDAVERAGE(*localDMABuffer,rightSample);
                        *localDMABuffer++ = (WORD)mixSample; /* Right DMA channel */

                        mixSample = SIGNEDWORDAVERAGE(*localDMABuffer,leftSample);
                        *localDMABuffer++ = (WORD)mixSample; /* Left DMA channel */

                        i --;
                    }

                    offset --;
                }
            }
            else
            {
    			offset = bufferUsed;
	       		while( offset )
		      	{
			     	currentSample = *localBuffer++;

				    i = dmaOutputReps;
				    while( i )
				    {
					   /* Right audio channel is first on Pluto, whereas left
					      is first in a WAV file. */
					   mixSample = SIGNEDWORDAVERAGE( *localDMABuffer, currentSample );
					   *localDMABuffer++ = mixSample; /* Right DMA channel */
					   *localDMABuffer++ = mixSample; /* Left DMA channel */
					   i --;
				    }

				    offset --;
                }
			}
			break;

		default:
			/* Incorrect channel so return */
			break;

	}

	localDMABufferUsed = localDMABuffer - localDMABufferStart;

	if ( wavDmaVars.bufferUsed[dmaChannel] < localDMABufferUsed )
	{
		wavDmaVars.bufferUsed[dmaChannel] = localDMABufferUsed;
	}
}


/*-------------------------------------------------------------------
Name......... FillDMABuffer()
Author....... Mark Perkins
Date......... 07/03/05
Description.. Fills the specified DMA buffer with data from any
............. currently playing WAV files. NOTE This function is
............. optimised for speed.
Parameters... Channel (either 0 or 1)
Return....... None
History...... 07/11/03 - Initial version  - MJB
............. 26/07/04 - Added (AA):
.............		U-law and A-law compressed file handling
............. 		byte swapping (endian conversion) now performed here
............. 07/03/05 - Changed to do fills over a number of interupts

-------------------------------------------------------------------*/
static void FillDMABuffer( UBYTE dmaChannel )
{
	ULONG	i;

	switch(fillstate)
	{
		case SNDSTATE_CHANNEL_LEFT0:
			ZeroDMABuffer(dmaChannel);

			/* Process LEFT0 or Stereo 0 Sample */
			ProcessAudioChannel(CHANNEL_LEFT0, dmaChannel);
			fillstate = SNDSTATE_CHANNEL_RIGHT0;
			break;

		case SNDSTATE_CHANNEL_RIGHT0:
			/* Process RIGHT0 or Stereo 0 Sample */
			ProcessAudioChannel(CHANNEL_RIGHT0, dmaChannel);
			fillstate = SNDSTATE_CHANNEL_LEFT1;
			break;

		case SNDSTATE_CHANNEL_LEFT1:
			/* Process LEFT1 or Stereo 1 Sample */
			ProcessAudioChannel(CHANNEL_LEFT1, dmaChannel);
			fillstate = SNDSTATE_CHANNEL_RIGHT1;
			break;

		case SNDSTATE_CHANNEL_RIGHT1:

			/* Process RIGHT1 or Stereo 1 Sample */
			ProcessAudioChannel(CHANNEL_RIGHT1, dmaChannel);

            /*
                Empire

                Stop Happens Sometine before, so bufferready gets set but
                no buffer contents
                DMA interrupt happens some time after test for DMA required
                but DMA channels stop because it checks
                the buffer has some contents but it does not so it sets
                DMA required flag.

                The next buffer is then filled but the DMA channels never
                get restarted as the DMA required flag is set for the other buffer.
            */
            if (wavDmaVars.bufferUsed[dmaChannel] > 0 )
            {
				wavDmaVars.bufferready[dmaChannel] = TRUE;
			}

			if ( wavDmaVars.dmaRequired[ dmaChannel ] )
			{
                /* What if interrupt here */
                if (dmaChannel == 0)
                {
          	        StartDMA2( (BYTE *)wavDmaVars.buffer[0], Sfxr1, (wavDmaVars.bufferUsed[0]*2) );
                    wavDmaVars.nextBufferToFill = 1;
                    wavDmaVars.bufferready[1] = FALSE;
                    wavDmaVars.bufferUsed[1] = 0;
                    wavDmaVars.activeDMAbuffer = 0;
        			wavDmaVars.bufferready[0] = FALSE;
                }
                else
                {
          	        StartDMA2( (BYTE *)wavDmaVars.buffer[1], Sfxr1, (wavDmaVars.bufferUsed[1]*2) );
                    wavDmaVars.nextBufferToFill = 0;
                    wavDmaVars.bufferready[0] = FALSE;
                    wavDmaVars.bufferUsed[0] = 0;
                    wavDmaVars.activeDMAbuffer = 1;
        			wavDmaVars.bufferready[1] = FALSE;
                 }

        		wavDmaVars.dmaRequired[dmaChannel] = FALSE;

            	for ( i = 0; i < NUM_CHANNELS; i ++ )
                {
                    if (wavFileChannel[i].status == CHANNEL_STOPING )
                    {
                        /* File Finished So Reset All Variables */
                        wavFileChannel[i].status = CHANNEL_IDLE;
                        wavFileChannel[i].channel = FALSE;
                        wavFileChannel[i].looped = FALSE;
                        wavFileChannel[i].fileLength = FALSE;
                        wavFileChannel[i].audioPos = FALSE;
                        wavFileChannel[i].numChannels = FALSE;
                        wavFileChannel[i].sampleRate = FALSE;
                        wavFileChannel[i].bitsPerSample = FALSE;
                        wavFileChannel[i].fileBufferPos = FALSE;
                        wavFileChannel[i].audioBegin = FALSE;
                        wavFileChannel[i].format = FALSE;

                    	FileRewind( &wavFileChannel[i].file );
                    }
               }
            }

			fillstate = SNDSTATE_CHANNEL_LEFT0;
			break;
	}
}


/*-------------------------------------------------------------------
Name......... ProcessAudioChannel()
Author....... Mark Perkins
Date......... 07/03/05
Description.. Process Audio Data For Given Channel.
............. process it. If there are any problems these will
............. be returned
Parameters... Channel to process, and dma channel
Return....... none
History...... 07/03/05 - Initial Version - MP.

-------------------------------------------------------------------*/
static void ProcessAudioChannel( UBYTE Channel, UBYTE dmaChannel)
{
	static WORD	tempBuffer[ WAV_DMA_BUFFER_SIZE ];
	static PROCESSBUFER	processBuffer;

	UBYTE	wavchannel;
	ULONG	wordsToRead=0, wordsRead=0;
	UBYTE	activeWavFile;
	UBYTE   errorCode;
	UBYTE	i;


	activeWavFile = FALSE;

	/* Start Playing a File */
	if (wavFileChannel[Channel].status == CHANNEL_START )
	{
		errorCode = ProcessWavHeader( Channel );
		if ( errorCode != 0 )
		{
			wavFileChannel[Channel].status = CHANNEL_WAVERROR;
			return;
		}

		if (
			/*
				Do some checking and only Play files which match the following channel option
					Channels Left0 - Mono or Stereo 0
					Channels Right0 - Mono or Stereo 0
					Channels Left1 - Mono or Stereo 1
					Channels Right1 - Mono or Stereo 1

					It is therefore not possible to play stereo files onto mono channels
			*/
			(Channel == CHANNEL_LEFT0 && wavFileChannel[CHANNEL_LEFT0].numChannels ==
			 STEREO && wavFileChannel[CHANNEL_LEFT0].channel != AUDIO_STEREO0) ||
			(Channel == CHANNEL_RIGHT0 && wavFileChannel[CHANNEL_RIGHT0].numChannels ==
			 STEREO && wavFileChannel[CHANNEL_LEFT0].channel != AUDIO_STEREO0) ||
			(Channel == CHANNEL_LEFT1 && wavFileChannel[CHANNEL_LEFT1].numChannels ==
			 STEREO && wavFileChannel[CHANNEL_LEFT1].channel != AUDIO_STEREO1) ||
			(Channel == CHANNEL_RIGHT1 && wavFileChannel[CHANNEL_RIGHT1].numChannels ==
			STEREO && wavFileChannel[CHANNEL_LEFT1].channel != AUDIO_STEREO1)
		)
		{
			wavFileChannel[Channel].status = CHANNEL_INVALIDCHANNEL;
			return;
		}
		else
		{
			wavFileChannel[Channel].status = CHANNEL_PLAY;
		}
	}

	if ( Channel == CHANNEL_RIGHT0 && wavFileChannel[CHANNEL_LEFT0].status ==
		CHANNEL_PLAY && wavFileChannel[CHANNEL_LEFT0].channel ==
		AUDIO_STEREO0 && wavFileChannel[CHANNEL_LEFT0].numChannels == STEREO)
	{
		/* Stereo fills are done over two interrupts */
		wavchannel = CHANNEL_LEFT0;
	}
	else if ( Channel == CHANNEL_RIGHT0 && wavFileChannel[CHANNEL_LEFT0].status ==
			 CHANNEL_PLAY && wavFileChannel[CHANNEL_LEFT0].channel ==
			 AUDIO_STEREO0 && wavFileChannel[CHANNEL_LEFT0].numChannels == MONO)
    {
        return;
    }
	else if (Channel == CHANNEL_RIGHT1 && wavFileChannel[CHANNEL_LEFT1].status ==
			 CHANNEL_PLAY && wavFileChannel[CHANNEL_LEFT1].channel ==
			 AUDIO_STEREO1 && wavFileChannel[CHANNEL_LEFT1].numChannels == STEREO)
	{
		/* Stereo fills are done over two interrupts */
		wavchannel = CHANNEL_LEFT1;
	}
	else if (Channel == CHANNEL_RIGHT1 && wavFileChannel[CHANNEL_LEFT1].status ==
			 CHANNEL_PLAY && wavFileChannel[CHANNEL_LEFT1].channel ==
			 AUDIO_STEREO1 && wavFileChannel[CHANNEL_LEFT1].numChannels == MONO)
    {
        return;
    }
	else
	{
		wavchannel = Channel;
	}

	if ( wavFileChannel[wavchannel].status == CHANNEL_PLAY )
	{

		/* We have at least one wav file currently playing */
		activeWavFile = TRUE;

		/* Assume 16kHz mono sample, so read 1 sector */
		wordsToRead = SAMPLES_REQUIRED;

		/* If a 32kHz sample, must read double samples (2 sectors) */
		if ( wavFileChannel[wavchannel].sampleRate == 32000 )
        {
        	wordsToRead <<= 1;
        }

		switch(wavFileChannel[wavchannel].format)
		{
			case FORMAT_PCM:
				if(wavFileChannel[wavchannel].bitsPerSample == 16)
				{
					wordsRead = LoadAudioData( wavchannel, processBuffer.bw, wordsToRead, 0 );
					ByteSwap(processBuffer.bw, tempBuffer, wordsRead);
				}
				else	/* bitsPerSample == 8 */
				{
					/* only need to read half the number of words */
					wordsRead = LoadAudioData( wavchannel, processBuffer.bw, wordsToRead/2, 1 );
					wordsRead <<=1;	/* doubled by conversion byte->word */
					UnsignedByte2SignedWord(processBuffer.bb, tempBuffer, wordsRead );
				}
				break;

			case FORMAT_ALAW:
				/* only need to read half the number of words */
				wordsRead = LoadAudioData( wavchannel, processBuffer.bw, wordsToRead/2, 1 );
				wordsRead <<=1;	/* doubled by decompression byte->word */
				AlawDecompress(processBuffer.bb, tempBuffer, wordsRead);
				break;

			case FORMAT_ULAW:
				/* only need to read half the number of words */
				wordsRead = LoadAudioData( wavchannel, processBuffer.bw, wordsToRead/2, 1 );
				wordsRead <<=1;	/* doubled by decompression byte->word */
				UlawDecompress(processBuffer.bb, tempBuffer, wordsRead);
				break;

			default:	/* should never happen */
				wordsRead=0;
				break;

		}

        FillDMASample( dmaChannel, wavchannel, tempBuffer, wordsRead );
	}

	activeWavFile = FALSE;
	for ( i = 0; i < NUM_CHANNELS; i ++ )
	{
		if (wavFileChannel[i].status == CHANNEL_START || wavFileChannel[i].status == CHANNEL_PLAY)
		{
			activeWavFile = TRUE;
		}
	}

	if (activeWavFile == 0 && wavFileChannel[wavchannel].status == CHANNEL_STOPING)
	{
		wavDmaVars.bufferUsed[dmaChannel] = wordsToRead;
	}

}


/*-------------------------------------------------------------------
Name......... ProcessWavHeader()
Author....... Mark Perkins
Date......... 07/03/06
Description.. Read the first sector of specified WAV file and
............. process it. If there are any problems these will
............. be returned
Parameters... wavFileInfo: Structure containing WAV file info
Return....... Errors, e.g. WAVERR_NORIFF
History...... 05/11/03 - Initial version  - MJB
............. 26/07/04 - Ulaw and Alaw compression detected - AA
............. 07/03/06 - Rewind file to make sure at start before doing anything.
-------------------------------------------------------------------*/
static BYTE ProcessWavHeader( UBYTE channel )
{
	UBYTE	fileSector[SECTOR_BYTES];
	UWORD	*infowordmemptr;
	ULONG	tempValue;			/* Temp value */
	ULONG	fmtLength;			/* fmt subchunk length (in bytes) */
	ULONG	fileOffset = 0;

	/* Always Rewind to start of file */
	FileRewind( &wavFileChannel[channel].file );

	wavFileChannel[channel].format=0;	/* !!!development: pcm uncompressed */

	/* Get Pointers to memory */
	(ULONG)infowordmemptr = (ULONG)&fileSector[ 0 ];

	/* Read First Sector as this contains WAV info */
	if ( ( FileGetSectors( currentDev, &wavFileChannel[channel].file, infowordmemptr, 1 ) ) == 1 )
	{
		/* Ensure that correct "RIFF" header is present. Return error if not. */
		if ( VerifyFileData( fileSector, &fileOffset, "RIFF", 4 ) != 0 )
		{
			return( WAVERR_NORIFF );
		}

		/* Obtain 4 BYTE (or LONG) file length (in little endian format) */
		wavFileChannel[channel].fileLength = GetFileNumber( fileSector, &fileOffset, 4 );

		/* Ensure that correct "WAVE" string is present. Return error if not. */
		if ( VerifyFileData( fileSector, &fileOffset, "WAVE", 4 ) != 0 )
		{
			return( WAVERR_NOWAVE );
		}

		/* Ensure that correct "fmt " string is present. Return error if not. */
		if ( VerifyFileData( fileSector, &fileOffset, "fmt ", 4 ) != 0 )
		{
			return( WAVERR_NOFMT );
		}

		/* Obtain "fmt" chunk length */
		fmtLength = GetFileNumber( fileSector, &fileOffset, 4 );

		/* Obtain audio format */
		wavFileChannel[channel].format = (WORD)GetFileNumber( fileSector, &fileOffset, 2 );
		if(( wavFileChannel[channel].format != FORMAT_PCM ) &&
           ( wavFileChannel[channel].format != FORMAT_ALAW ) &&
           ( wavFileChannel[channel].format != FORMAT_ULAW ))
        {
			return(WAVERR_UNSUPPORTEDFMT);
        }

		/* Obtain number of channels, e.g. 1 = mono, 2 = stereo etc */
		wavFileChannel[channel].numChannels = (WORD)GetFileNumber( fileSector, &fileOffset, 2 );
		if ( ( wavFileChannel[channel].numChannels != 1 ) && ( wavFileChannel[channel].numChannels != 2 ) )
        {
            return( WAVERR_CHANNELS );
        }

		/* Obtain sample rate (must be equal to either 16000 or 32000) */
		wavFileChannel[channel].sampleRate = GetFileNumber( fileSector, &fileOffset, 4 );
		if ( ( wavFileChannel[channel].sampleRate != 16000 ) && ( wavFileChannel[channel].sampleRate != 32000 ) )
        {
            return( WAVERR_SAMPLERATE );
        }

		/* UNUSED : Byte rate = SampleRate * NumChannels * BitsPerSample/8 */
		tempValue = GetFileNumber( fileSector, &fileOffset, 4 );

		/* UNUSED : Block align = NumChannels * BitsPerSample/8. The number of
           bytes for one sample including all channels. */
		tempValue = GetFileNumber( fileSector, &fileOffset, 2 );

		/* Obtain bits per sample, e.g. 8, 16. */
		wavFileChannel[channel].bitsPerSample = (WORD)GetFileNumber( fileSector, &fileOffset, 2 );

		/* Skip remainder of "fmt " section */
		fileOffset = fmtLength + 20;

		/* Find the "data" section. If there are any other sections before, skip them. */
		while( VerifyFileData( fileSector, &fileOffset, "data", 4 ) != 0 )
		{
			/* Obtain chunk length */
			tempValue = GetFileNumber( fileSector, &fileOffset, 4 );
			fileOffset += tempValue;
		}

		/* Obtain audio data length (in WORDS) */
		wavFileChannel[channel].audioLength = GetFileNumber( fileSector, &fileOffset, 4 ) / 2;

		/* Sample audio data follows... */
	}
	else
    {
        return( WAVERR_READ );
    }

	/* Convert file offset from BYTES to WORDS (this will always work as the
       offset should always be an even number) */
	wavFileChannel[channel].audioBegin = fileOffset / 2;

	ResetAudioFile( channel );

	return( 0 );
}



/*-------------------------------------------------------------------
Name......... UlawDecompress
Author....... Alan Austin
Date......... 26 July 2004
Description.. Decompress U-law samples using look-up table
............. Optimised for speed (fastest generated asm code)
Parameters... Pointer to compressed source bytes
............. Pointer to decompressed destination words
............. number of samples
Return....... none
History...... 26/07/2004 Compressed sound initial version
-------------------------------------------------------------------*/
static void UlawDecompress(UBYTE *srcb, WORD *dstw, ULONG nsamples)
{
	UBYTE *end;

	end = srcb + nsamples;

	do {
		*dstw++ = ulaw_lookup[*srcb++];
		*dstw++ = ulaw_lookup[*srcb++];
	} while (srcb < end);

	/* 'while' loop is faster than 'for' loop - no decrementing loop counter needed */

}

/*-------------------------------------------------------------------
Name......... AlawDecompress
Author....... Alan Austin
Date......... 26 July 2004
Description.. Decompress A-law samples using look-up table
............. Optimised for speed (fastest generated asm code)
Parameters... Pointer to compressed source bytes
............. Pointer to decompressed destination words
............. number of samples
Return....... none
History...... 26/07/2004 Compressed sound initial version
-------------------------------------------------------------------*/
static void AlawDecompress(UBYTE *srcb, WORD *dstw, ULONG nsamples)
{
	UBYTE *end;

	end = srcb + nsamples;

	do {
		*dstw++ = alaw_lookup[*srcb++];
		*dstw++ = alaw_lookup[*srcb++];
	} while (srcb < end);

	/* 'while' loop is faster than 'for' loop - no decrementing loop counter needed */

}

/*-------------------------------------------------------------------
Name......... UnsignedByte2SignedWord
Author....... Alan Austin
Date......... 26 July 2004
Description.. Convert wav file data samples: 8-bit to 16-bit
.............
............. Optimised for speed (fastest generated asm code)
Parameters... Pointer to source bytes
............. Pointer to destination words
............. number of samples
Return....... none
History...... 26/07/2004 Compressed sound initial version
-------------------------------------------------------------------*/
static void UnsignedByte2SignedWord(UBYTE *srcb, WORD *dstw, ULONG nsamples)
{
	UBYTE *end;

	end = srcb + nsamples;

	/*
	8-bit format: unsigned byte
	16-bit format: signed word
	*/
	do {
		*dstw++ = 0x8000 + (*srcb++ <<8);
		*dstw++ = 0x8000 + (*srcb++ <<8);
	} while (srcb < end);

	/* 'while' loop is faster than 'for' loop - no decrementing loop counter needed */

}


/*-------------------------------------------------------------------
Name......... ByteSwap
Author....... Mark Perkins
Date......... 07/03/05
Description.. Endian conversion: wav file data -> Pluto 6
.............
............. Optimised for speed by Mark Perkins 15th Febuary 2004
............. Optimised for speed (fastest generated asm code)
Parameters... Pointer to wav file format words
............. Pointer to byte swapped destination words
............. number of samples
Return....... none
History...... 26/07/2004 Compressed sound initial version
............. 07/03/05 - Rewrite MP -- Rewrite to speed up.
-------------------------------------------------------------------*/
static void ByteSwap(UWORD *srcw, UWORD *dstw,  ULONG nsamples)
{
	register UBYTE	*srclsb, *srcmsb;
	register UBYTE	*dstlsb, *dstmsb;
	register ULONG	samples;
	UWORD	srcsample;
	UWORD	dstsample;

	(ULONG)srclsb = (ULONG)&srcsample;
	(ULONG)srcmsb = (ULONG)&srcsample;
	srcmsb++;

	(ULONG)dstlsb = (ULONG)&dstsample;
	(ULONG)dstmsb = (ULONG)&dstsample;
	dstmsb++;

	samples = nsamples;
	while(samples)
	{
		srcsample = *srcw++;

		*dstlsb = *srcmsb;
		*dstmsb = *srclsb;

		*dstw++ = dstsample;

		samples--;
	}
}

/**********************************************************************************************/
/*                                                                                            */
/* File access functions                                                                      */
/*                                                                                            */
/**********************************************************************************************/

/*-------------------------------------------------------------------
Name......... ResetAudioFile()
Author....... Mike Ballinger
Date......... 11/12/03
Description.. Resets specified WAV file.
Parameters... wavFileInfo: Structure containing WAV file info
Return....... Nothing
History...... 11/12/03 - Initial version  - MJB
-------------------------------------------------------------------*/
static void ResetAudioFile( UBYTE	channel )
{
	UWORD	*infowordmemptr;

	FileRewind( &wavFileChannel[channel].file );

	/* Get Pointers to memory */
	(ULONG)infowordmemptr = (ULONG)&wavFileChannel[channel].fileBuffer[0];

	/* Read First Sector as this contains WAV info */
	FileGetSectors( currentDev, &wavFileChannel[channel].file, infowordmemptr, 1 );
	wavFileChannel[channel].BufferEnd = SECTOR_WORDS;

	wavFileChannel[channel].fileBufferPos = wavFileChannel[channel].audioBegin;
	wavFileChannel[channel].audioPos = 0;
}

/*-------------------------------------------------------------------
Name......... LoadAudioData()
Author....... Mark Perkins
Date......... 07/03/06
Description.. Loads the specified number of words from audio file.
............. NOTE: This function is optimised for speed
Parameters... wavFileInfo: Structure containing WAV file info
............. buffer: Output buffer
............. wordsToRead: Number of words to read
Return....... Number of words read
History...... 11/12/03 - Initial version  - MJB
............. 07/03/05 - Rewrite MP -- Main change is to load a differing
............. number of sectors depending on format.
-------------------------------------------------------------------*/
static LONG LoadAudioData( UBYTE channel, WORD *buffer, ULONG wordsToRead, UBYTE compressed )
{
	register ULONG wordsLeftToRead;
	register ULONG localFileBufferPos;
	register ULONG localAudioPos;
	register ULONG localAudioLength;
	UWORD *infowordmemptr;
	UWORD *localBuffer, *localFileBuffer;
	ULONG filebufferend;
	ULONG	sectorsread;

	/* Make local copies of
	variables to speed up function */
	localAudioPos = wavFileChannel[channel].audioPos;
	localAudioLength = wavFileChannel[channel].audioLength;
	localBuffer = buffer;
	localFileBufferPos = wavFileChannel[channel].fileBufferPos;
	localFileBuffer = &wavFileChannel[channel].fileBuffer[localFileBufferPos];

	/* Get Pointer to file buffer for FileGetSectors() */
	(ULONG)infowordmemptr = (ULONG)&wavFileChannel[channel].fileBuffer[0];

	filebufferend = wavFileChannel[channel].BufferEnd;

	wordsLeftToRead = wordsToRead;
	while( wordsLeftToRead )
	{
		if ( localAudioPos >= localAudioLength )
		{
		/* We have reached the end of the WAV file. If looping, continue
               from beginning of WAV file, otherwise leave function. */
			if ( wavFileChannel[channel].looped )
			{
				wavFileChannel[channel].status = CHANNEL_START;
				return ( wordsToRead - wordsLeftToRead );
			}
			else
			{
				wavFileChannel[channel].status = CHANNEL_STOPING;
				return ( wordsToRead - wordsLeftToRead );
			}

		}
		else if ( localFileBufferPos >= filebufferend )
		{
            /* We have reached the end of the buffer so load some more data */
			if (compressed)
			{
				if (wavFileChannel[channel].sampleRate == 32000 )
				{
					sectorsread = FileGetSectors( currentDev, &wavFileChannel[channel].file,
												 infowordmemptr, CACHE_SECTORSCOMPRESSED32K );
				}
				else
				{
					sectorsread = FileGetSectors( currentDev, &wavFileChannel[channel].file,
					 							 infowordmemptr, CACHE_SECTORSCOMPRESSED );
				}
			}
			else
			{
				if (wavFileChannel[channel].sampleRate == 32000 )
				{
					sectorsread = FileGetSectors( currentDev, &wavFileChannel[channel].file,
					infowordmemptr, CACHE_SECTORS32K );
				}
				else
				{
					sectorsread = FileGetSectors( currentDev, &wavFileChannel[channel].file,
					infowordmemptr, CACHE_SECTORS );
				}
			}

			if (sectorsread)
			{
				wavFileChannel[channel].BufferEnd = (sectorsread * SECTOR_WORDS);
				filebufferend = wavFileChannel[channel].BufferEnd;

				/* Reset buffer position counter */
				localFileBufferPos = 0;
				localFileBuffer = &wavFileChannel[channel].fileBuffer[localFileBufferPos];
			}
			else
            {
				if ( wavFileChannel[channel].looped )
				{
					wavFileChannel[channel].status = CHANNEL_START;
					return ( wordsToRead - wordsLeftToRead );
				}
				else
				{
					wavFileChannel[channel].status = CHANNEL_STOPING;
					return ( wordsToRead - wordsLeftToRead );
				}
            }
		}

		*localBuffer++ = *localFileBuffer++;
		localFileBufferPos++;
		localAudioPos++;
		wordsLeftToRead--;
	}

	/* Update global variables with local copies */
	wavFileChannel[channel].audioPos = localAudioPos;
	wavFileChannel[channel].fileBufferPos = localFileBufferPos;

	return ( wordsToRead - wordsLeftToRead );
}

/*-------------------------------------------------------------------
Name......... ClearAudioChannel()
Author....... Mark Perkins
Date......... 22/04/05
Description.. Clear's audio channel structure.
.............
Parameters... audioChannel: channel, e.g. AUDIO_STEREO
Return....... None
History...... 19/08/2004 - Initial version - AA
............. 07/03/05 - Rewrite MP -- Changed to stop channel by using the status.
-------------------------------------------------------------------*/
static void ClearAudioChannel( UBYTE audioChannel )
{
	wavFileChannel[audioChannel].status = CHANNEL_IDLE;
	wavFileChannel[audioChannel].channel = FALSE;
	wavFileChannel[audioChannel].looped = FALSE;
    wavFileChannel[audioChannel].fileLength = FALSE;
    wavFileChannel[audioChannel].audioPos = FALSE;
    wavFileChannel[audioChannel].numChannels = FALSE;
    wavFileChannel[audioChannel].sampleRate = FALSE;
    wavFileChannel[audioChannel].bitsPerSample = FALSE;
    wavFileChannel[audioChannel].fileBufferPos = FALSE;
    wavFileChannel[audioChannel].audioBegin = FALSE;
    wavFileChannel[audioChannel].format = FALSE;
    FileRewind( &wavFileChannel[audioChannel].file );
}

/**********************************************************************************************/
/*                                                                                            */
/* Interrupt Functions                                                                        */
/*                                                                                            */
/**********************************************************************************************/
/*-------------------------------------------------------------------
Name......... SFX10msInterruptRoutine()
Author....... Mark Perkins
Date......... 07/03/06
Description.. 10ms interrupt. If any sound files are currently
............. playing, ensure that both DMA buffers are filled.
Parameters... None
Return....... None
History...... 07/11/03 - Initial version - MJB
............. 07/03/05 - Rewrite MP -- Changed SFX Interrupt rotuine
............. to fill buffer over a number of interrupts.
-------------------------------------------------------------------*/
static void SFX10msInterruptRoutine()
{
	register ULONG	i;
	register ULONG	activeWavFile;

	/* Already In Here so dont do it again */
	if (In10ms) return;
	In10ms = TRUE;

	activeWavFile = FALSE;
	for ( i = 0; i < NUM_CHANNELS; i ++ )
	{
		if (wavFileChannel[i].status == CHANNEL_START || wavFileChannel[i].status == CHANNEL_PLAY)
		{
			activeWavFile = TRUE;
		}
	}

	if (activeWavFile)
	{
		/* If either DMA buffer is empty then fill it. */
		if ( ( wavDmaVars.bufferready[0] == 0 ) && ( wavDmaVars.nextBufferToFill == 0 ) )
		{
			if (fillstate == SNDSTATE_CHANNEL_RIGHT1)
			{
				wavDmaVars.nextBufferToFill = 1;
			}
			FillDMABuffer( 0 );
		}
		else if ( ( wavDmaVars.bufferready[1] == 0 ) && ( wavDmaVars.nextBufferToFill == 1 ) )
		{
			if (fillstate == SNDSTATE_CHANNEL_RIGHT1)
			{
				wavDmaVars.nextBufferToFill = 0;
			}
			FillDMABuffer( 1 );
		}
	}

	In10ms = FALSE;
}

/*-------------------------------------------------------------------
Name......... DMA2InterruptRoutine()
Author....... Mark Perkins
Date......... 07/03/06
Description.. DMA2 interrupt service function (interrupt on DMA
............. transfer complete)
............. Swap DMA buffers to enable playback to continue
Parameters... None
Return....... None
History...... 17/11/03 - Initial version - MJB
............. 07/03/05 - Rewrite MP -- Only DMA buffer's which are ready.
-------------------------------------------------------------------*/
static void DMA2InterruptRoutine()
{
    register ULONG	i;

    /* Stop DMA Channel */
	StopDMA2();

    /* Swap DMA Buffer if buffer is ready */
    if ( wavDmaVars.activeDMAbuffer == 1 )
    {
		/* buffer 1 is no longer valid - refill next 10ms interrupt */
        wavDmaVars.bufferready[1] = FALSE;
        wavDmaVars.bufferUsed[1] = 0;
        wavDmaVars.activeDMAbuffer = 0;

        if( wavDmaVars.bufferUsed[0] > 0 && wavDmaVars.bufferready[0] )
        {
             /* Start DMA Channel with Buffer 0 */
	         StartDMA2( (BYTE *)wavDmaVars.buffer[0], Sfxr1, (wavDmaVars.bufferUsed[0]*2) );
        }
        else
        {
             /* DMA buffer not ready so start in the next ready */
	        wavDmaVars.dmaRequired[0] = TRUE;
        }
	}
	else
    {
  		/* buffer 0 is no longer valid - refill next 10ms interrupt */
        wavDmaVars.bufferready[0] = FALSE;
        wavDmaVars.bufferUsed[0] = 0;
        wavDmaVars.activeDMAbuffer = 1;

        if( wavDmaVars.bufferUsed[1] > 0 && wavDmaVars.bufferready[1] )
        {
             /* Start DMA Channel with Buffer 1 */
    	    StartDMA2( (BYTE *)wavDmaVars.buffer[1], Sfxr1, (wavDmaVars.bufferUsed[1]*2) );
       }
        else
        {
             /* DMA buffer not ready so start in the next ready */
        	wavDmaVars.dmaRequired[1] = TRUE;
        }
    }

    /* Check for channels finishing */
	for ( i = 0; i < NUM_CHANNELS; i ++ )
    {
        if (wavFileChannel[i].status == CHANNEL_STOPING )
        {
            /* File Finished So Reset All Variables */
            wavFileChannel[i].status = CHANNEL_IDLE;
            wavFileChannel[i].channel = FALSE;
            wavFileChannel[i].looped = FALSE;
            wavFileChannel[i].fileLength = FALSE;
            wavFileChannel[i].audioPos = FALSE;
            wavFileChannel[i].numChannels = FALSE;
            wavFileChannel[i].sampleRate = FALSE;
            wavFileChannel[i].bitsPerSample = FALSE;
            wavFileChannel[i].fileBufferPos = FALSE;
            wavFileChannel[i].audioBegin = FALSE;
            wavFileChannel[i].format = FALSE;

        	FileRewind( &wavFileChannel[i].file );
        }
    }
}
